"""
Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

  You might be tempted to import things from __main__ later, but that will cause
  problems: the code will get executed twice:

  - When you run `python -mholdup` python will execute
    ``__main__.py`` as a script. That means there won't be any
    ``holdup.__main__`` in ``sys.modules``.
  - When you import __main__ it will get executed again (as a module) because
    there's no ``holdup.__main__`` in ``sys.modules``.

  Also see (1) from http://click.pocoo.org/5/setuptools/#setuptools-integration
"""
from __future__ import print_function

import argparse
import ast
import os
import socket
import ssl
import sys
from contextlib import closing
from time import sleep
from time import time

try:
    from pipes import quote
except ImportError:
    from shlex import quote

try:
    import builtins
except ImportError:
    import __builtin__ as builtins

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    from inspect import getargspec

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


class Check(object):
    error = None

    def is_passing(self, options):
        try:
            self.run(options)
        except Exception as exc:
            self.error = exc
        else:
            if options.verbose:
                print('holdup: Passed check: %r' % self)
            return True

    def __repr__(self):
        if self.error:
            return '{0} ({0.error})'.format(self)
        else:
            return str(self)


class TcpCheck(Check):
    def __init__(self, host, port):
        self.host = host
        self.port = port

    def run(self, options):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(options.check_timeout)
        with closing(sock):
            sock.connect((self.host, self.port))

    def __str__(self):
        return 'tcp://{0.host}:{0.port}'.format(self)


class HttpCheck(Check):
    def __init__(self, url):
        self.do_insecure = False
        proto, urn = url.split('://', 1)
        if proto == 'https+insecure':
            self.do_insecure = True
            proto = 'https'
        self.url = '{}://{}'.format(proto, urn)

    def can_create_default_context(self):
        """
            Check if the current python version supports
                * ssl.create_default_context()
                * 'context' kwargs for urlopen
            Supported Python versions are:
                * >2.7.9
                * >3.4.3
        """
        if hasattr(ssl, 'create_default_context'):
            urlopen_argspec = getargspec(urlopen)
            urlopen_args = urlopen_argspec.args
            if hasattr(urlopen_argspec, 'kwonlyargs'):
                urlopen_args.extend(urlopen_argspec.kwonlyargs)
            if 'context' in urlopen_args:
                return True
            else:
                return False
        else:
            return False

    def run(self, options):
        kwargs = {}
        do_insecure = self.do_insecure
        if options.insecure:
            do_insecure = True
        if self.can_create_default_context():
            ssl_ctx = ssl.create_default_context()
            if do_insecure:
                ssl_ctx.check_hostname = False
                ssl_ctx.verify_mode = ssl.CERT_NONE
            kwargs = {'context': ssl_ctx}
        else:
            if do_insecure:
                raise Exception("Insecure HTTPS is not supported with the current version of python")
        with closing(urlopen(self.url, timeout=options.check_timeout, **kwargs)) as req:
            status = req.getcode()
            if status != 200:
                raise Exception("Expected status code 200, got: %r." % status)

    def __str__(self):
        return self.url


class UnixCheck(Check):
    def __init__(self, path):
        self.path = path

    def run(self, options):
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(options.check_timeout)
        with closing(sock):
            sock.connect(self.path)

    def __str__(self):
        return 'unix://{0.path}'.format(self)


class PathCheck(Check):
    def __init__(self, path):
        self.path = path

    def run(self, _):
        os.stat(self.path)
        if not os.access(self.path, os.R_OK):
            raise Exception("Failed access(%r, 'R_OK') test." % self.path)

    def __str__(self):
        return 'path://{0.path}'.format(self)


class EvalCheck(Check):
    def __init__(self, expr):
        self.expr = expr
        self.ns = {}
        try:
            tree = ast.parse(expr)
        except SyntaxError as exc:
            raise argparse.ArgumentTypeError('Invalid service spec %r. Parse error:\n'
                                             '  %s %s^\n'
                                             '%s' % (expr, exc.text, ' '*exc.offset, exc))
        for node in ast.walk(tree):
            if isinstance(node, ast.Name):
                if not hasattr(builtins, node.id):
                    try:
                        __import__(node.id)
                    except ImportError as exc:
                        raise argparse.ArgumentTypeError('Invalid service spec %r. Import error: %s' % (expr, exc))
                    self.ns[node.id] = sys.modules[node.id]

    def run(self, _):
        result = eval(self.expr, dict(self.ns), dict(self.ns))
        if not result:
            raise Exception("Failed to evaluate %r. Result %r is falsey." % (self.expr, result))

    def __str__(self):
        return 'eval://{0.expr}'.format(self)


class AnyCheck(Check):
    def __init__(self, checks):
        self.checks = checks

    def run(self, options):
        errors = []
        for check in self.checks:
            if check.is_passing(options):
                return
            else:
                errors.append(check)
        if errors:
            raise Exception("Nothing succeeded: %s." % ', '.join(repr(check) for check in errors))

    def __str__(self):
        return 'any(%s)' % ','.join(str(check) for check in self.checks)


def parse_service(service):
    if '://' not in service:
        raise argparse.ArgumentTypeError('Invalid service spec %r. Must have "://".' % service)
    proto, value = service.split('://', 1)

    if ',' in value and proto != 'eval':
        parts = value.split(',')
        for pos, part in enumerate(parts):
            if part.startswith('eval://'):
                parts[pos] = ','.join(parts[pos:])
                del parts[pos + 1:]
                break
        return AnyCheck([parse_value(part, proto) for part in parts])
    else:
        return parse_value(value, proto)


def parse_value(value, proto):
    if '://' in value:
        proto, value = value.split('://', 1)

    if proto == 'tcp':
        if ':' not in value:
            raise argparse.ArgumentTypeError('Invalid service spec %r. Must have ":". Where\'s the port?' % value)
        host, port = value.strip('/').split(':', 1)
        if not port.isdigit():
            raise argparse.ArgumentTypeError('Invalid service spec %r. Port must be a number.' % value)
        port = int(port)
        return TcpCheck(host, port)
    elif proto == 'unix':
        return UnixCheck(value)
    elif proto == 'path':
        return PathCheck(value)
    elif proto in ('http', 'https', 'https+insecure'):
        return HttpCheck('%s://%s' % (proto, value))
    elif proto == 'eval':
        return EvalCheck(value)
    else:
        raise argparse.ArgumentTypeError('Unknown protocol %r in %r. Must be "tcp", "path" or "unix".' % (proto, value))


parser = argparse.ArgumentParser(
    usage='%(prog)s [-h] [-t SECONDS] [-T SECONDS] [-i SECONDS] [-n] service [service ...] [-- command [arg [arg ...]]]',
    description="Wait for services to be ready and optionally exec command."
)
parser.add_argument('service', nargs=argparse.ONE_OR_MORE, type=parse_service,
                    help='A service to wait for. '
                         'Supported protocols: "tcp://host:port/", "path:///path/to/something", '
                         '"unix:///path/to/domain.sock", "eval://expr", '
                         '"http://urn", "http://urn", "https+insecure//urn" (status 200 expected). '
                         'Join protocols with a comma to make holdup exit at the first '
                         'passing one, eg: tcp://host:1,host:2 or tcp://host:1,tcp://host:2 are equivalent and mean '
                         '"any that pass".')
parser.add_argument('command', nargs=argparse.OPTIONAL,
                    help='An optional command to exec.')
parser.add_argument('-t', '--timeout', metavar='SECONDS', type=float, default=60.0,
                    help='Time to wait for services to be ready. Default: %(default)s')
parser.add_argument('-T', '--check-timeout', metavar='SECONDS', type=float, default=1.0,
                    help='Time to wait for a single check. Default: %(default)s')
parser.add_argument('-i', '--interval', metavar='SECONDS', type=float, default=.2,
                    help='How often to check. Default: %(default)s')
parser.add_argument('-v', '--verbose', action='store_true',
                    help='Verbose mode. ')
parser.add_argument('-n', '--no-abort', action='store_true',
                    help='Ignore failed services. '
                         'This makes `holdup` return 0 exit code regardless of services actually responding.')
parser.add_argument('--insecure', action='store_true',
                    help='Disable SSL Certificate verification for HTTPS services')


def main():
    """
    Args:
        argv (list): List of arguments

    Returns:
        int: A return code

    Does stuff.
    """
    if '--' in sys.argv:
        pos = sys.argv.index('--')
        argv, command = sys.argv[1:pos], sys.argv[pos + 1:]
    else:
        argv, command = sys.argv[1:], None
    options = parser.parse_args(args=argv)
    if options.timeout < options.check_timeout:
        if options.check_timeout == 1.0:
            options.check_timeout = options.timeout
        else:
            parser.error('--timeout value must be greater than --check-timeout value!')
    pending = list(options.service)
    if options.verbose:
        print('holdup: Waiting for %ss (%ss per check, %ss sleep between loops) for these services: %s' % (
            options.timeout, options.check_timeout, options.interval, ' '.join(map(repr, pending))))
    start = time()
    at_least_once = True
    while at_least_once or pending and time() - start < options.timeout:
        lapse = time()
        pending = [check for check in pending if not check.is_passing(options)]
        sleep(max(0, options.interval - time() + lapse))
        at_least_once = False

    if pending:
        if options.no_abort:
            print('holdup: Failed checks:', ', '.join(repr(check) for check in pending), file=sys.stderr)
        else:
            parser.exit(1, 'holdup: Failed service checks: %s. Aborting!\n' % ', '.join(repr(check) for check in pending))
    if command:
        if options.verbose:
            print('holdup: Executing: %s' % ' '.join(map(quote, command)))
        os.execvp(command[0], command)



import os
import socket
import ssl
import threading

import pytest

try:
    from inspect import getfullargspec as getargspec
except ImportError:
    from inspect import getargspec

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

pytest_plugins = 'pytester',


def urlopen_has_ssl_context():
    if hasattr(ssl, 'create_default_context'):
        urlopen_argspec = getargspec(urlopen)
        urlopen_args = urlopen_argspec.args
        if hasattr(urlopen_argspec, 'kwonlyargs'):
            urlopen_args.extend(urlopen_argspec.kwonlyargs)
        if 'context' in urlopen_args:
            return False
        else:
            return True
    else:
        return True


@pytest.fixture(params=[[], ['--', 'python', '-c', 'print("success !")']])
def extra(request):
    return request.param


def test_normal(testdir, extra):
    tcp = socket.socket()
    tcp.bind(('127.0.0.1', 0))
    tcp.listen(1)
    _, port = tcp.getsockname()
    t = threading.Thread(target=tcp.accept)
    t.start()

    uds = socket.socket(socket.AF_UNIX)
    if os.path.exists('/tmp/holdup-test.sock'):
        os.unlink('/tmp/holdup-test.sock')
    with open('/tmp/holdup-test', 'w'):
        pass
    uds.bind('/tmp/holdup-test.sock')
    uds.listen(1)
    result = testdir.run(
        'holdup',
        '-t', '0.5',
        'tcp://localhost:%s/' % port,
        'path:///tmp/holdup-test',
        'unix:///tmp/holdup-test.sock',
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines(['success !'])
    assert result.ret == 0
    t.join()


@pytest.mark.parametrize('status', [200, 404])
@pytest.mark.parametrize('proto', ['http', 'https'])
def test_http(testdir, extra, status, proto):
    result = testdir.run(
        'holdup',
        '-T', '5',
        '-t', '5.1',
        '%s://httpbin.org/status/%s' % (proto, status),
        *extra
    )
    if extra:
        if status == 200:
            result.stdout.fnmatch_lines(['success !'])
        else:
            result.stderr.fnmatch_lines(['*HTTP Error 404*'])


@pytest.mark.skipif("urlopen_has_ssl_context()")
def test_http_insecure_with_option(testdir):
    result = testdir.run(
        'holdup',
        '-t', '2',
        '--insecure',
        'https://self-signed.badssl.com/',
    )
    assert result.ret == 0


@pytest.mark.skipif("urlopen_has_ssl_context()")
def test_http_insecure_with_proto(testdir):
    result = testdir.run(
        'holdup',
        '-t', '2',
        'https+insecure://self-signed.badssl.com/',
    )
    assert result.ret == 0


def test_any(testdir, extra):
    tcp = socket.socket()
    tcp.bind(('127.0.0.1', 0))
    _, port = tcp.getsockname()

    uds = socket.socket(socket.AF_UNIX)
    if os.path.exists('/tmp/holdup-test.sock'):
        os.unlink('/tmp/holdup-test.sock')
    uds.bind('/tmp/holdup-test.sock')
    uds.listen(1)
    result = testdir.run(
        'holdup',
        '-v',
        '-t', '0.5',
        'tcp://localhost:%s/,path:///tmp/holdup-test,unix:///tmp/holdup-test.sock' % port,
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines([
            'holdup: Waiting for 0.5s (0.5s per check, 0.2s sleep between loops) for these services: '
            'any(tcp://localhost:*,path:///tmp/holdup-test,unix:///tmp/holdup-test.sock)',
            'holdup: Passed check: path:///tmp/holdup-test',
            'holdup: Passed check: any(tcp://localhost:*,path:///tmp/holdup-test,unix:///tmp/holdup-test.sock)',
            'holdup: Executing: python -c \'print("success !")\'',
            'success !'
        ])
    assert result.ret == 0


def test_any_same_proto(testdir, extra):
    tcp1 = socket.socket()
    tcp1.bind(('127.0.0.1', 0))
    _, port1 = tcp1.getsockname()

    tcp2 = socket.socket()
    tcp2.bind(('127.0.0.1', 0))
    tcp2.listen(1)
    _, port2 = tcp2.getsockname()
    t = threading.Thread(target=tcp2.accept)
    t.start()

    result = testdir.run(
        'holdup',
        '-t', '0.5',
        'tcp://localhost:%s,localhost:%s/' % (port1, port2),
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines(['success !'])
    assert result.ret == 0
    t.join()


def test_any_failed(testdir):
    tcp = socket.socket()
    tcp.bind(('127.0.0.1', 0))
    _, port = tcp.getsockname()

    result = testdir.run(
        'holdup',
        '-t', '0.5',
        'tcp://localhost:%s/,path:///doesnt/exist,unix:///doesnt/exist' % port,
    )
    result.stderr.fnmatch_lines([
        'holdup: Failed service checks: any(tcp://localhost:%s,path:///doesnt/exist,unix:///doesnt/exist) '
        '(Nothing succeeded: '
        'tcp://localhost:%s (*), '
        'path:///doesnt/exist (*), '
        'unix:///doesnt/exist (*). Aborting!' % (port, port)
    ])


def test_no_abort(testdir, extra):
    result = testdir.run(
        'holdup',
        '-t', '0.1',
        '-n',
        'tcp://localhost:0',
        'tcp://localhost:0/',
        'path:///doesnt/exist',
        'unix:///doesnt/exist',
        *extra
    )
    result.stderr.fnmatch_lines([
        'holdup: Failed checks: tcp://localhost:0 (*), '
        'path:///doesnt/exist (*), unix:///doesnt/exist (*)'
    ])


@pytest.mark.skipif(os.path.exists('/.dockerenv'), reason="chmod(0) does not work in docker")
def test_not_readable(testdir, extra):
    foobar = testdir.maketxtfile(foobar='')
    foobar.chmod(0)
    result = testdir.run(
        'holdup',
        '-t', '0.1',
        '-n',
        'path://%s' % foobar,
        *extra
    )
    result.stderr.fnmatch_lines(["holdup: Failed checks: path://%s (Failed access('%s', 'R_OK') test.)" % (foobar, foobar)])


def test_bad_timeout(testdir):
    result = testdir.run(
        'holdup',
        '-t', '0.1',
        '-T', '2',
        'path:///'
    )
    result.stderr.fnmatch_lines([
        '*error: --timeout value must be greater than --check-timeout value!'
    ])


def test_eval_bad_import(testdir):
    result = testdir.run(
        'holdup',
        'eval://foobar123.foo()'
    )
    result.stderr.fnmatch_lines([
        "*error: argument service: Invalid service spec 'foobar123.foo()'. Import error: No module named*"
    ])


def test_eval_bad_expr(testdir):
    result = testdir.run(
        'holdup',
        'eval://foobar123.foo(.)'
    )
    result.stderr.fnmatch_lines([
        "*error: argument service: Invalid service spec 'foobar123.foo(.)'. Parse error:",
        '  foobar123.foo(.)',
        '*               ^',
        'invalid syntax (<unknown>, line 1)',
    ])


def test_eval_bad_pg(testdir):
    pytest.importorskip('psycopg2')
    result = testdir.run(
        'holdup',
        '-t', '0.1',
        'eval://psycopg2.connect("dbname=foo host=0.0.0.0")'
    )
    result.stderr.fnmatch_lines([
        'holdup: Failed service checks: eval://psycopg2.connect* (*'
    ])


def test_eval_falsey(testdir):
    result = testdir.run(
        'holdup',
        '-t', '0',
        'eval://None'
    )
    result.stderr.fnmatch_lines([
        "holdup: Failed service checks: eval://None (Failed to evaluate 'None'. Result None is falsey.). Aborting!"
    ])
    assert result.ret == 1


def test_eval_distutils(testdir, extra):
    result = testdir.run(
        'holdup',
        'eval://distutils.spawn.find_executable("find")',
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines(['success !'])
    assert result.ret == 0


def test_eval_comma(testdir, extra):
    result = testdir.run(
        'holdup',
        'eval://os.path.join("foo", "bar")',
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines(['success !'])
    assert result.ret == 0


def test_eval_comma_anycheck(testdir, extra):
    result = testdir.run(
        'holdup',
        'path://whatever123,eval://os.path.join("foo", "bar")',
        *extra
    )
    if extra:
        result.stdout.fnmatch_lines(['success !'])
    assert result.ret == 0
